import logging
from collections import defaultdict

from pandas import DataFrame

from .abstract import AbstractFeatureGenerator
from ..feature_metadata import R_INT, R_FLOAT

logger = logging.getLogger(__name__)


# TODO: Not necessary to exist after fitting, can just update outer context feature_out/feature_in and then delete this
class DropDuplicatesFeatureGenerator(AbstractFeatureGenerator):
    """Drops features which are exact duplicates of other features, leaving only one instance of the data."""
    def _fit_transform(self, X: DataFrame, **kwargs) -> (DataFrame, dict):
        features_to_drop = self._drop_duplicate_features(X)
        self._remove_features_in(features_to_drop)
        if features_to_drop:
            self._log(15, f'\t{len(features_to_drop)} duplicate columns removed: {features_to_drop}')
        X_out = X[self.features_in]
        return X_out, self.feature_metadata_in.type_group_map_special

    def _transform(self, X: DataFrame) -> DataFrame:
        return X

    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict()

    # TODO: optimize categorical/object handling
    # TODO: Potentially use subsamples?
    def _drop_duplicate_features(self, X: DataFrame) -> list:
        feature_sum_map = defaultdict(list)
        for feature in self.feature_metadata_in.get_features(valid_raw_types=[R_INT, R_FLOAT]):
            feature_sum_map[round(X[feature].sum(), 2)].append(feature)

        features_to_keep = []
        features_to_remove = []
        for key in feature_sum_map:
            if len(feature_sum_map[key]) <= 1:
                features_to_keep += feature_sum_map[key]
            else:
                features_to_keep += list(X[feature_sum_map[key]].T.drop_duplicates().T.columns)
                features_to_remove += [feature for feature in feature_sum_map[key] if feature not in features_to_keep]

        if features_to_keep or features_to_remove:
            X = X.drop(columns=features_to_keep + features_to_remove)

        if len(X.columns) > 0:
            X_without_dups = X.T.drop_duplicates().T
            columns_orig = X.columns.values
            columns_new = X_without_dups.columns.values
            features_to_remove += [column for column in columns_orig if column not in columns_new]

        return features_to_remove

    def _more_tags(self):
        return {'feature_interactions': False}



import random
from calendar import monthrange
from datetime import datetime
from functools import reduce

import numpy as np
from pytz import timezone

TZ = timezone("Asia/Seoul")


class LogicProcessor:
    """
    상품 선택 로직 등을 프로세스하는 클래스입니다.

    아래 operations를 참조하여 json 형식의 로직을 구현하면 됩니다.

    ## 기본 문법

    기본적으로 JSON 형태의 로직은 아래 포맷을 따릅니다.

    ```python
    {
        "operator": [*args]
    }
    ```

    각각의 arg는 또 다른 JSON 형태의 로직이 될 수 있으며 recursive하게 프로세스됩니다.

    ```json
    {
        "and": [
            {"==": [3, 3]},
            {">": [
                5,
                {"max": [3, 10]}
            ]}
        ]
    }
    ```

    ## 피쳐 참조

    피쳐를 참조하는 방법은

    ```
    {
        "var": ["feature_name", 기본값]
    }
    ```

    형식입니다.

    리스트 또는 딕셔너리 타입의 피쳐 값을 참조하는 경우

    ```
    {
        "var": ["feature_name.0", 기본값]
    }
    ```

    ```
    {
        "var": ["feature_name.key_name", 기본값]
    }
    ```

    과 같은 형태로 하위 값에 접근이 가능합니다.

    ## 추가 키 참조

    Recommendation API v3 이후부터 지원하는 추가 키(세컨 키 등) 리스트는

    ```
    {
        "var": ["additional_keys.0", 기본값]
    }
    ```

    와 같이 `additional_key.index` 포맷으로 접근 가능합니다.

    ## 연산자

    - `==`: 두 값이 같은 경우 True. 타입을 고려하지 않음
    - `===`: 두 값이 같은 경우 True. 타입을 고려함
    - `!=`: 두 값이 다른 경우 True. 타입을 고려하지 않음
    - `!==`: 두 값이 다른 경우 True. 타입을 고려함
    - `<`: 왼쪽 값이 오른쪽 값보다 작은 경우 True
    - `<=`: 왼쪽 값이 오른쪽 값보다 작거나 같은 경우 True
    - `>`: 왼쪽 값이 오른쪽 값보다 큰 경우 True
    - `>=`: 왼쪽 값이 오른쪽 값보다 크거나 같은 경우 True
    - `!`: not
    - `and`: 모든 args가 True인 경우 True
    - `or`: 하나 이상의 args가 True인 경우 True
    - `?:`: args[0]이 True이면 args[1] 반환, 아니면 args[2] 반환
    - `if`: args[0]이 True이면 args[1] 반환, 아닌 경우 만약 args[2]이 True이면 args[3] 반환, ... , 조건이 모두 False면 args[-1] 반환
    - `if_min`: args[0], args[2], args[4]... 중 args[i]가 최소인 경우 args[i + 1] 반환
    - `if_max`: args[0], args[2], args[4]... 중 args[i]가 최대인 경우 args[i + 1] 반환
    - `if_closest`: args[1], args[3], args[5]... 중 args[0]과 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `if_closest_upper`: args[1], args[3], args[5]... 중 args[0]보다 크면서 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `if_closest_lower`: args[1], args[3], args[5]... 중 args[0]보다 작으면서 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `in`: 왼쪽 값이 오른쪽 값에 포함된 경우 True
    - `cat`: args들을 문자열 join
    - `+`: args들 합
    - `-`: 왼쪽 빼기 오른쪽. args의 길이가 1인 경우 -args[0]
    - `/`: 왼쪽 나누기 오른쪽
    - `%`: 왼쪽 나누기 오른쪽한 나머지
    - `min`: args 중 최소값
    - `max`: args 중 최대값
    - `argmin`: 최소값을 가지는 인덱스
    - `argmax`: 최대값을 가지는 인덱스
    - `get`: args[0]의 args[1]번째 element. 딕셔너리 타입의 경우 args[0]의 args[1] 키에 해당하는 값
    - `int`: int 파싱
    - `float`: float 파싱
    - `str`: str 파싱
    - `list`: args들을 리스트에 넣어 그대로 반환
    - `dict`: {args[0]: args[1], args[2]: args[3], ...}
    - `dict_keys`: 딕셔너리 타입의 args[0]의 키 리스트 반환
    - `dict_values`: 딕셔너리 타입의 args[0]의 값 리스트 반환
    - `pick`: args[0] 리스트 element 중 args[1]에서 정의한 인덱스만 뽑아냄
    - `pick_in`: args[0] 리스트 element 중 args[1]에 존재하는 것들만 필터
    - `remove`: args[1]에서 args[0] 값을 가진 element들을 삭제 후 반환
    - `remove_empty`: args[0]에서 빈 자료구조나 None인 element들을 삭제
    - `sample`: args[0] 중 args[1]개 만큼 랜덤 추출 (seed: args[2])
    - `first`: args[0] 중 첫 args[1]개를 반환
    - `replace`: args[2]의 element 중 args[0] 값을 args[1] 값으로 교체하여 반환
    - `repeat`: args[0]이 args[1]번 반복되는 리스트 반환
    - `len`: args[0]의 길이 반환
    - `empty`: args[0]의 길이가 0이면 True
    - `unique`: args[0] 리스트에서 중복값을 제거
    - `merge`: args들을 하나의 리스트로 합침
    - `vmerge`: 길이가 같은 리스트 args[0], args[1], args[2]... 의 같은 순서 element들끼리 묶어 반환
    - `count`: args의 길이
    - `abs`: args[0]의 절대값
    - `closest`: args[1], args[2]... 중 args[0]과 가장 가까운 값
    - `closest_upper`: args[1], args[2]... 중 args[0]보다 크면서 가장 가까운 값
    - `closest_lower`: args[1], args[2]... 중 args[0]보다 작으면서 가장 가까운 값
    - `weekday`: 요일 인덱스 반환 (월요일 1 ~ 일요일 7)
    - `ndays`: 이번 달의 마지막 날짜
    - `day`: 오늘 날짜
    - `month`: 이번 달
    - `year`: 올해
    - `hour`: 시
    - `minute`: 분
    - `second`: 초
    - `microsecond`: 마이크로초
    - `rand`: 0~1 사이의 난수 (seed: args[0])
    - `pf`: 프로파일 API 조회. (api_type: args[0], profile_id: args[1], item_id/user_id: args[2], keys: args[3])
    - `with`: args[-1]의 프로세스에 {args[0]: arg[1], args[2]: args[3]...}의 데이터를 `var`로 참조 가능.
    - `for`: args[0]를 iterator로 사용하여 args[1] 루프. {"var": "i"}, {"var": "i.0"} 등으로 iterator 참조 가능 (map과 유사)
    """

    def __init__(self):
        self.operations = {
            "==": self._soft_equals,
            "===": self._hard_equals,
            "!=": lambda a, b: not self._soft_equals(a, b),
            "!==": lambda a, b: not self._hard_equals(a, b),
            ">": lambda a, b: self._less(b, a),
            ">=": lambda a, b: self._less(b, a) or self._soft_equals(a, b),
            "<": self._less,
            "<=": self._less_or_equal,
            "!": lambda a: not a,
            "!!": bool,
            "and": lambda *args: reduce(lambda total, arg: total and arg, args, True),
            "or": lambda *args: reduce(lambda total, arg: total or arg, args, False),
            "?:": lambda a, b, c: b if a else c,
            "if": self._if,
            "if_min": self._if_min,
            "if_max": self._if_max,
            "if_closest": self._if_closest,
            "if_closest_upper": self._if_closest_upper,
            "if_closest_lower": self._if_closest_lower,
            "in": lambda a, b: a in b if "__contains__" in dir(b) else False,
            "contains": lambda a, b: b in a if "__contains__" in dir(a) else False,
            "cat": lambda *args: "".join(str(arg) for arg in args),
            "+": self._plus,
            "*": lambda *args: reduce(lambda total, arg: total * float(arg), args, 1),
            "-": self._minus,
            "/": lambda a, b=None: a if b is None else float(a) / float(b),
            "%": lambda a, b: a % b,
            "min": lambda *args: min(args),
            "max": lambda *args: max(args),
            "argmin": lambda *args: np.argmin(args),
            "argmax": lambda *args: np.argmax(args),
            "get": self._get,
            "int": self._int,
            "float": self._float,
            "str": self._str,
            "list": lambda *args: list(args),
            "dict": self._dict,
            "dict_keys": lambda a: list(a.keys()),
            "dict_values": lambda a: list(a.values()),
            "pick": lambda a, b: [a[i] for i in b],
            "pick_in": self._pick_in,
            "remove": lambda a, b: [i for i in b if i != a],
            "remove_empty": lambda a: [i for i in a if i],
            "sample": self._sample,
            "first": lambda a, num_pick: list(a[:num_pick]),
            "replace": lambda a, b, c: [b if i == a else i for i in c],
            "repeat": lambda a, n: [a] * n,
            "len": lambda a: len(a),
            "empty": lambda a: len(a) == 0,
            "unique": self._unique,
            "merge": self._merge,
            "vmerge": self._vmerge,
            "count": lambda *args: sum(1 if a else 0 for a in args),
            "abs": lambda a: abs(a),
            "closest": self._closest,
            "closest_upper": self._closest_upper,
            "closest_lower": self._closest_lower,
            "weekday": lambda: datetime.now(TZ).isoweekday(),
            "ndays": lambda: monthrange(datetime.now(TZ).year, datetime.now(TZ).month)[1],
            "day": lambda: datetime.now(TZ).day,
            "month": lambda: datetime.now(TZ).month,
            "year": lambda: datetime.now(TZ).year,
            "hour": lambda: datetime.now(TZ).hour,
            "minute": lambda: datetime.now(TZ).minute,
            "second": lambda: datetime.now(TZ).second,
            "microsecond": lambda: datetime.now(TZ).microsecond,
            "rand": self._rand,
            "return": lambda *args: args,
        }

    def _if_min(self, *args):
        min_arg = np.argmin([args[i] for i in range(0, len(args) - 1, 2)])
        return args[min_arg * 2 + 1]

    def _if_max(self, *args):
        max_arg = np.argmax([args[i] for i in range(0, len(args) - 1, 2)])
        return args[max_arg * 2 + 1]

    def _if_closest(self, arg, *comps):
        args = list(comps)
        for i in range(0, len(args) - 1, 2):
            args[i] = abs(arg - args[i])
        return self._if_min(*args)

    def _if_closest_upper(self, arg, *comps):
        args = []
        for i in range(0, len(comps) - 1, 2):
            if comps[i] > arg:
                args.append(comps[i])
                args.append(comps[i + 1])
        if not args:
            return None
        return self._if_closest(arg, *args)

    def _if_closest_lower(self, arg, *comps):
        args = []
        for i in range(0, len(comps) - 1, 2):
            if comps[i] < arg:
                args.append(comps[i])
                args.append(comps[i + 1])
        if not args:
            return None
        return self._if_closest(arg, *args)

    def _soft_equals(self, a, b):
        if isinstance(a, str) or isinstance(b, str):
            return str(a) == str(b)
        if isinstance(a, bool) or isinstance(b, bool):
            return bool(a) is bool(b)
        return a == b

    def _hard_equals(self, a, b):
        if type(a) != type(b):
            return False
        return a == b

    def _less(self, a, b, *args):
        types = set([type(a), type(b)])
        if float in types or int in types:
            try:
                a, b = float(a), float(b)
            except TypeError:
                return False
        return a < b and (not args or self._less(b, *args))

    def _less_or_equal(self, a, b, *args):
        return (self._less(a, b) or self._soft_equals(a, b)) and (not args or self._less_or_equal(b, *args))

    def _to_numeric(self, arg):
        if isinstance(arg, str):
            if "." in arg:
                return float(arg)
            else:
                return int(arg)
        return arg

    def _plus(self, *args):
        return sum(self._to_numeric(arg) for arg in args)

    def _minus(self, *args):
        if len(args) == 1:
            return -self._to_numeric(args[0])
        return self._to_numeric(args[0]) - self._to_numeric(args[1])

    def _closest(self, arg, *comps):
        closest = comps[0]
        min_diff = abs(arg - comps[0])
        for i in range(1, len(comps)):
            diff = abs(arg - comps[i])
            if min_diff > diff:
                min_diff = diff
                closest = comps[i]
        return closest

    def _closest_upper(self, arg, *comps):
        args = [c for c in comps if c > arg]
        if not args:
            return None
        return self._closest(arg, *args)

    def _closest_lower(self, arg, *comps):
        args = [c for c in comps if c < arg]
        if not args:
            return None
        return self._closest(arg, *args)

    def _get(self, a, b, not_found=None):
        try:
            return a[b]
        except (KeyError, TypeError, ValueError, IndexError):
            return not_found

    def _float(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [float(i) for i in a]
        return float(a)

    def _int(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [int(i) for i in a]
        return int(a)

    def _str(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [str(i) for i in a]
        return str(a)

    def _dict(self, *args):
        d = {}
        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            d[args[i]] = args[i + 1]
        return d

    def _pick_in(self, a, b):
        result = []
        for i in a:
            if i in b:
                result.append(i)
        return result

    def _sample(self, a, num_pick, seed):
        random.seed(seed)
        return random.sample(a, num_pick)

    def _rand(self, seed):
        random.seed(seed)
        return random.random()

    def _unique(self, a):
        result = []
        for item in a:
            if item not in result:
                result.append(item)
        return result

    def _merge(self, *args):
        ret = []
        for arg in args:
            if isinstance(arg, list) or isinstance(arg, tuple):
                ret += list(arg)
            else:
                ret.append(arg)
        return ret

    def _vmerge(self, *args):
        ret = []
        for i in range(len(args[0])):
            r = []
            for j in range(len(args)):
                r.append(args[j][i])
            ret.append(r)
        return ret

    def _pf(self, pf_client, *args):
        if not pf_client:
            return None

        try:
            api_type = args[0]
            profile_id = args[1]
            target_id = args[2]
            keys = args[3]
        except (KeyError, TypeError, ValueError, IndexError):
            return None

        if api_type == "user":
            pf = pf_client.get_user_profile(profile_id=profile_id, user_id=target_id, keys=keys)
        else:
            pf = pf_client.get_item_profile(profile_id=profile_id, item_id=target_id, keys=keys)

        return [pf.get(key) for key in keys]

    def _if(self, data, pf_client, *args):
        for i in range(0, len(args) - 1, 2):
            if self.apply(args[i], data, pf_client):
                return self.apply(args[i + 1], data, pf_client)
        if len(args) % 2:
            return self.apply(args[-1], data, pf_client)
        else:
            return None

    def _with(self, data, pf_client, *args):
        if len(args) % 2 == 0:
            raise ValueError("with operator에 오류가 있습니다.")

        body = args[-1]
        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            data[args[i]] = self.apply(args[i + 1], data, pf_client)

        results = self.apply(body, data, pf_client)

        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            data.pop(args[i], None)

        return results

    def _for(self, data, pf_client, *args):
        results = []
        it = self.apply(args[0], data, pf_client)
        for i in it:
            data["i"] = i
            results.append(self.apply(args[1], data, pf_client))
        data.pop("i", None)
        return results

    def _get_var(self, data, var_name, not_found=None):
        try:
            for key in str(var_name).split("."):
                try:
                    data = data[key]
                except TypeError:
                    data = data[int(key)]
        except (KeyError, TypeError, ValueError):
            return not_found
        else:
            return data

    def _missing(self, data, *args):
        not_found = object()
        if args and isinstance(args[0], list):
            args = args[0]
        ret = []
        for arg in args:
            if self._get_var(data, arg, not_found) is not_found:
                ret.append(arg)
        return ret

    def _missing_some(self, data, min_required, args):
        if min_required < 1:
            return []
        found = 0
        not_found = object()
        ret = []
        for arg in args:
            if self._get_var(data, arg, not_found) is not_found:
                ret.append(arg)
            else:
                found += 1
                if found >= min_required:
                    return []
        return ret

    def apply(self, tests, data=None, pf_client=None):
        """
        ## Args

        - tests: (dict) 테스트 할 json 형식의 로직
        - data: (optional) (dict) 로직에서 참조할 데이터 dict (기본값: None)
        - pf_client: (optional) (`sktmls.apis.profile_api.MLSProfileAPIClient`) 로직에서 사용할 프로파일 API 클라이언트 (기본값: None)
        """
        if tests is None or not isinstance(tests, dict):
            return tests

        operator = list(tests.keys())[0]
        data = data or {}
        values = tests[operator]

        if operator == "if":
            return self._if(data, pf_client, *values)

        if operator == "with":
            return self._with(data, pf_client, *values)

        if operator == "for":
            return self._for(data, pf_client, *values)

        if not isinstance(values, list) and not isinstance(values, tuple):
            values = [values]

        values = [self.apply(val, data, pf_client) for val in values]

        if operator == "pf":
            return self._pf(pf_client, *values)
        if operator == "var":
            return self._get_var(data, *values)
        if operator == "missing":
            return self._missing(data, *values)
        if operator == "missing_some":
            return self._missing_some(data, *values)

        if operator not in self.operations:
            raise ValueError(f"없는 operator입니다: {operator}")

        return self.operations[operator](*values)



# Could eventually remove this code: Is this needed in unit tests?

"""
Object definitions that are used for testing.
"""

from typing import Iterator
import numpy as np

from ..datatypes.common import StateIdAndCandidate
from ..datatypes.hp_ranges import HyperparameterRanges_Impl, HyperparameterRangeContinuous, HyperparameterRangeInteger, HyperparameterRangeCategorical
from ..datatypes.scaling import LogScaling, LinearScaling
from ..tuning_algorithms.base_classes import CandidateGenerator
from ..tuning_algorithms.default_algorithm import dictionarize_objective


class RepeatedCandidateGenerator(CandidateGenerator):
    """Generates candidates from a fixed set. Used to test the deduplication logic."""
    def __init__(self, n_unique_candidates: int):
        self.all_unique_candidates = [
            (1.0*j, j, "value_" + str(j))
            for j in range(n_unique_candidates)
        ]

    def generate_candidates(self) -> Iterator[StateIdAndCandidate]:
        i = 0
        while True:
            i += 1
            yield self.all_unique_candidates[i % len(self.all_unique_candidates)]


# Example black box function, with adjustable location of global minimum.
# Potentially could catch issues with optimizer, e.g. if the optimizer
# ignoring somehow candidates on the edge of search space.
# A simple quadratic function is used.
class Quadratic3d:
    def __init__(self, local_minima, active_metric, metric_names):
        # local_minima: point where local_minima is located
        self.local_minima = np.array(local_minima).astype('float')
        self.local_minima[0] = np.log10(self.local_minima[0])
        self.active_metric = active_metric
        self.metric_names = metric_names

    @property
    def search_space(self):
        return HyperparameterRanges_Impl(
            HyperparameterRangeContinuous('x', 1.0, 100.0, scaling=LogScaling()),
            HyperparameterRangeInteger('y', 0, 2, scaling=LinearScaling()),
            HyperparameterRangeCategorical('z', ('0.0', '1.0', '2.0'))
        )

    @property
    def f_min(self):
        return 0.0

    def __call__(self, candidate):
        p = np.array([float(hp) for hp in candidate])
        p[0] = np.log10(p[0])
        return dictionarize_objective(np.sum((self.local_minima - p) ** 2))



import copy
import inspect
import logging
from collections import defaultdict

from sklearn.linear_model import LogisticRegression, LinearRegression

from ...constants import AG_ARGS, AG_ARGS_FIT, BINARY, MULTICLASS, REGRESSION, SOFTCLASS, PROBLEM_TYPES_CLASSIFICATION
from ...models.abstract.abstract_model import AbstractModel
from ...models.fastainn.tabular_nn_fastai import NNFastAiTabularModel
from ...models.lgb.lgb_model import LGBModel
from ...models.lr.lr_model import LinearModel
from ...models.tabular_nn.tabular_nn_model import TabularNeuralNetModel
from ...models.rf.rf_model import RFModel
from ...models.knn.knn_model import KNNModel
from ...models.catboost.catboost_model import CatboostModel
from ...models.xt.xt_model import XTModel
from ....metrics import soft_log_loss, mean_squared_error

logger = logging.getLogger(__name__)

# Higher values indicate higher priority, priority dictates the order models are trained for a given level.
DEFAULT_MODEL_PRIORITY = dict(
    RF=100,
    XT=90,
    KNN=80,
    GBM=70,
    CAT=60,
    NN=50,
    FASTAI=45,
    LR=40,
    custom=0,
)

# Problem type specific model priority overrides (will update default values in DEFAULT_MODEL_PRIORITY)
PROBLEM_TYPE_MODEL_PRIORITY = {
    MULTICLASS: dict(
        NN=120,
        FASTAI=115,
        KNN=110,
    ),
}

DEFAULT_SOFTCLASS_PRIORITY = dict(
    GBM=100,
    NN=90,
    RF=80,
    CAT=60,
    custom=0,
)

DEFAULT_CUSTOM_MODEL_PRIORITY = 0

MODEL_TYPES = dict(
    RF=RFModel,
    XT=XTModel,
    KNN=KNNModel,
    GBM=LGBModel,
    CAT=CatboostModel,
    NN=TabularNeuralNetModel,
    LR=LinearModel,
    FASTAI=NNFastAiTabularModel,
)

DEFAULT_MODEL_NAMES = {
    RFModel: 'RandomForest',
    XTModel: 'ExtraTrees',
    KNNModel: 'KNeighbors',
    LGBModel: 'LightGBM',
    CatboostModel: 'Catboost',
    TabularNeuralNetModel: 'NeuralNet',
    LinearModel: 'LinearModel',
    NNFastAiTabularModel: 'FastAINeuralNet',
}


def _dd_classifier():
    return 'Classifier'


def _dd_regressor():
    return 'Regressor'


DEFAULT_MODEL_TYPE_SUFFIX = dict(
    classifier=defaultdict(_dd_classifier),
    regressor=defaultdict(_dd_regressor),
)
DEFAULT_MODEL_TYPE_SUFFIX['classifier'].update({LinearModel: ''})
DEFAULT_MODEL_TYPE_SUFFIX['regressor'].update({LinearModel: ''})


# DONE: Add levels, including 'default'
# DONE: Add lists
# DONE: Add custom which can append to lists
# DONE: Add special optional AG args for things like name prefix, name suffix, name, etc.
# TODO: Move creation of stack ensemble internally into this function? Requires passing base models in as well.
# DONE: Add special optional AG args for training order
# TODO: Add special optional AG args for base models
# TODO: Consider making hyperparameters arg in fit() accept lists, concatenate hyperparameter sets together.
# TODO: Consider adding special optional AG args for #cores,#gpus,num_early_stopping_iterations,etc.
# DONE: Consider adding special optional AG args for max train time, max memory size, etc.
# TODO: Consider adding special optional AG args for use_original_features,features_to_use,etc.
# TODO: Consider adding optional AG args to dynamically disable models such as valid_num_classes_range, valid_row_count_range, valid_feature_count_range, etc.
# TODO: Args such as max_repeats, num_folds
# TODO: Add banned_model_types arg
# TODO: Add option to update hyperparameters with only added keys, so disabling CatBoost would just be {'CAT': []}, which keeps the other models as is.
# TODO: special optional AG arg for only training model if eval_metric in list / not in list. Useful for F1 and 'is_unbalanced' arg in LGBM.
def get_preset_models(path, problem_type, eval_metric, hyperparameters, stopping_metric=None, num_classes=None, hyperparameter_tune=False,
                      level='default', extra_ag_args_fit=None, name_suffix='', default_priorities=None):
    if problem_type not in [BINARY, MULTICLASS, REGRESSION, SOFTCLASS]:
        raise NotImplementedError
    if default_priorities is None:
        default_priorities = copy.deepcopy(DEFAULT_MODEL_PRIORITY)
        if problem_type in PROBLEM_TYPE_MODEL_PRIORITY:
            default_priorities.update(PROBLEM_TYPE_MODEL_PRIORITY[problem_type])

    if level in hyperparameters.keys():
        level_key = level
    else:
        level_key = 'default'
    hp_level = hyperparameters[level_key]
    priority_dict = defaultdict(list)
    for model_type in hp_level:
        for model in hp_level[model_type]:
            model = copy.deepcopy(model)
            if AG_ARGS not in model:
                model[AG_ARGS] = dict()
            if 'model_type' not in model[AG_ARGS]:
                model[AG_ARGS]['model_type'] = model_type
            model_priority = model[AG_ARGS].get('priority', default_priorities.get(model_type, DEFAULT_CUSTOM_MODEL_PRIORITY))
            # Check if model is valid
            if hyperparameter_tune and model[AG_ARGS].get('disable_in_hpo', False):
                continue  # Not valid
            priority_dict[model_priority].append(model)
    model_priority_list = [model for priority in sorted(priority_dict.keys(), reverse=True) for model in priority_dict[priority]]
    model_names_set = set()
    models = []
    for model in model_priority_list:
        model_type = model[AG_ARGS]['model_type']
        if not inspect.isclass(model_type):
            model_type = MODEL_TYPES[model_type]
        elif not issubclass(model_type, AbstractModel):
            logger.warning(f'Warning: Custom model type {model_type} does not inherit from {AbstractModel}. This may lead to instability. Consider wrapping {model_type} with an implementation of {AbstractModel}!')
        else:
            logger.log(20, f'Custom Model Type Detected: {model_type}')
        name_orig = model[AG_ARGS].get('name', None)
        if name_orig is None:
            name_main = model[AG_ARGS].get('name_main', DEFAULT_MODEL_NAMES.get(model_type, model_type.__name__))
            name_prefix = model[AG_ARGS].get('name_prefix', '')
            name_type_suffix = model[AG_ARGS].get('name_type_suffix', None)
            if name_type_suffix is None:
                suffix_key = 'classifier' if problem_type in (PROBLEM_TYPES_CLASSIFICATION+[SOFTCLASS]) else 'regressor'
                name_type_suffix = DEFAULT_MODEL_TYPE_SUFFIX[suffix_key][model_type]
            name_suff = model[AG_ARGS].get('name_suffix', '')
            name_orig = name_prefix + name_main + name_type_suffix + name_suff
        name = name_orig
        num_increment = 2
        while name in model_names_set:  # Ensure name is unique
            name = f'{name_orig}_{num_increment}'
            num_increment += 1
        model_names_set.add(name)
        model_params = copy.deepcopy(model)
        model_params.pop(AG_ARGS)
        if extra_ag_args_fit is not None:
            if AG_ARGS_FIT not in model_params:
                model_params[AG_ARGS_FIT] = {}
            model_params[AG_ARGS_FIT].update(extra_ag_args_fit.copy())  # TODO: Consider case of overwriting user specified extra args.
        model_init = model_type(path=path, name=name, problem_type=problem_type, eval_metric=eval_metric, stopping_metric=stopping_metric, num_classes=num_classes, hyperparameters=model_params)
        models.append(model_init)

    for model in models:
        model.rename(model.name + name_suffix)

    return models


def get_preset_stacker_model(path, problem_type, eval_metric, num_classes=None,
                             hyperparameters={'NN': {}, 'GBM': {}}, hyperparameter_tune=False):
    # TODO: Expand options to RF and NN
    if problem_type == REGRESSION:
        model = RFModel(path=path, name='LinearRegression', model=LinearRegression(),
                        problem_type=problem_type, eval_metric=eval_metric)
    else:
        model = RFModel(path=path, name='LogisticRegression', model=LogisticRegression(
            solver='liblinear', multi_class='auto', max_iter=500,  # n_jobs=-1  # TODO: HP set to hide warnings, but we should find optimal HP for this
        ), problem_type=problem_type, eval_metric=eval_metric)
    return model


def get_preset_models_softclass(path, hyperparameters, num_classes=None, hyperparameter_tune=False, name_suffix=''):
    model_types_standard = ['GBM','NN','CAT']
    hyperparameters = copy.deepcopy(hyperparameters)
    hyperparameters_standard = copy.deepcopy(hyperparameters)
    hyperparameters_rf = copy.deepcopy(hyperparameters)
    default_level_key = 'default'
    if default_level_key in hyperparameters:
        hyperparameters_standard[default_level_key] = {key: hyperparameters_standard[default_level_key][key] for key in hyperparameters_standard[default_level_key] if key in model_types_standard}
        hyperparameters_rf[default_level_key] = {key: hyperparameters_rf[default_level_key][key] for key in hyperparameters_rf[default_level_key] if key == 'RF'}
    else:
        hyperparameters_standard ={key: hyperparameters_standard[key] for key in hyperparameters_standard if key in model_types_standard}
        hyperparameters_rf ={key: hyperparameters_rf[key] for key in hyperparameters_rf if key == 'RF'}
        # TODO: add support for per-stack level hyperparameters
    models = get_preset_models(path=path, problem_type=SOFTCLASS, eval_metric=soft_log_loss, stopping_metric=soft_log_loss,
                               hyperparameters=hyperparameters_standard, num_classes=num_classes, hyperparameter_tune=hyperparameter_tune,
                                name_suffix=name_suffix, default_priorities=DEFAULT_SOFTCLASS_PRIORITY)
    # Swap RF criterion for MSE:
    rf_models = []
    if len(hyperparameters_rf) > 0:
        rf_newparams = {'criterion': 'mse', 'AG_args': {'name_suffix': 'MSE'}}
        if 'RF' in hyperparameters_rf:
            rf_params = hyperparameters_rf['RF']
        elif 'default' in hyperparameters_rf and 'RF' in hyperparameters_rf['default']:
            rf_params = hyperparameters_rf['default']['RF']
        else:
            rf_params = None
        if isinstance(rf_params, list):
            for i in range(len(rf_params)):
                rf_params[i].update(rf_newparams)
            rf_params = [j for n, j in enumerate(rf_params) if j not in rf_params[(n+1):]]  # Remove duplicates which may arise after overwriting criterion
        elif rf_params is not None:
            rf_params.update(rf_newparams)
        if 'RF' in hyperparameters_rf:
            hyperparameters_rf['RF'] = rf_params
        elif 'default' in hyperparameters_rf and 'RF' in hyperparameters_rf['default']:
            hyperparameters_rf['default']['RF'] = rf_params
        rf_models = get_preset_models(path=path, problem_type=REGRESSION, eval_metric=mean_squared_error,
                                      hyperparameters=hyperparameters_rf, hyperparameter_tune=hyperparameter_tune,
                                      name_suffix=name_suffix, default_priorities=DEFAULT_SOFTCLASS_PRIORITY)
    models_cat = [model for model in models if 'Catboost' in model.name]
    models_noncat = [model for model in models if 'Catboost' not in model.name]
    models = models_noncat + rf_models + models_cat
    if len(models) == 0:
        raise ValueError("At least one of the following model-types must be present in hyperparameters: ['GBM','CAT','NN','RF'], "
                         "These are the only supported models for softclass prediction problems. "
                         "Softclass problems are also not yet supported for fit() with per-stack level hyperparameters.")
    for model in models:
        model.normalize_pred_probas = True
        model.name = model.name.replace('Regressor', 'Classifier')  # conceal from user that model may actually be a regressor.

    return models


"""
def get_preset_models_softclassOLD(path, num_classes=None, hyperparameters={'GBM':{}, 'CAT':{}, 'NN':{}, 'RF':{}},
                                hyperparameter_tune=False, name_suffix=''):
    models = []
    extra_models = []
    model_types_standard = ['GBM','NN','CAT']
    for modeltype_name in model_types_standard:
        modeltype_options = None
        if modeltype_name in hyperparameters:
            modeltype_options = hyperparameters[modeltype_name]
        elif 'default' in hyperparameters and modeltype_name in hyperparameters['default']:
            modeltype_options = hyperparameters['default'][modeltype_name]
        if modeltype_options is not None:
            model_class = MODEL_TYPES[modeltype_name]
            display_name = DEFAULT_MODEL_NAMES[model_class] + 'SoftClassifier'
            if not isinstance(modeltype_options, list):
                modeltype_options = [modeltype_options]
            for modeltype_option in modeltype_options:
                model_obj = model_class(path=path, name=display_name, problem_type=SOFTCLASS, num_classes=num_classes,
                                        eval_metric=soft_log_loss, stopping_metric=soft_log_loss, hyperparameters=modeltype_option.copy())
                if modeltype_name == 'CAT':
                    extra_models.append(model_obj)  # only train CAT models last
                else:
                    models.append(model_obj)

    rf_type_name = 'RF'
    rf_options = hyperparameters.get(rf_type_name, None)
    rf_params = {'model_type': 'rf'}
    rf_params.update(hyperparameters.copy())
    rf_params['criterion'] = 'mse'
    if rf_options is not None:
        rf_class = MODEL_TYPES[rf_type_name]
        rf_display_name = DEFAULT_MODEL_NAMES[rf_class] + 'SoftClassifier'
        models.append(rf_class(path=path, name=rf_display_name, problem_type=REGRESSION,
                               eval_metric=mean_squared_error, hyperparameters=rf_params)
        )

    models = models + extra_models
    for model in models:
        model.rename(model.name + name_suffix)
        model.normalize_pred_probas = True

    return models
"""


from typing import Callable, Tuple, List
import numpy as np

from . import SliceException

MAX_STEP_OUT = 200
MAX_STEP_LOOP = 200


class SliceSampler(object):

    def __init__(self, log_density: Callable[[np.ndarray], float], scale: float, random_seed: int):
        self.log_density = log_density
        self.scale = scale  # default in scala core is 1.0
        self.random_state = np.random.RandomState(random_seed)

    def _gen_next_sample(self, x0: np.ndarray) -> np.ndarray:
        random_direction = gen_random_direction(len(x0), self.random_state)

        def sliced_log_density(_movement: float) -> float:
            return self.log_density(x0 + random_direction * _movement)

        # a quantity used to determine the bounds and accept movement along random_direction
        log_pivot = sliced_log_density(0.0) + np.log(self.random_state.rand())

        lower_bound, upper_bound = slice_sampler_step_out(log_pivot, self.scale, sliced_log_density, self.random_state)
        movement = slice_sampler_step_in(lower_bound, upper_bound, log_pivot, sliced_log_density, self.random_state)
        return x0 + random_direction * movement

    def sample(self, init_sample: np.ndarray, num_samples: int, burn: int, thin: int) -> List[np.ndarray]:
        samples = []
        next_sample = init_sample
        for _ in range(num_samples):
            next_sample = self._gen_next_sample(next_sample)
            samples.append(next_sample)
        return samples[burn::thin]


def gen_random_direction(dimension: int, random_state: np.random.RandomState) -> np.ndarray:
    random_direction = random_state.randn(dimension)
    random_direction *= 1.0 / np.linalg.norm(random_direction)
    return random_direction


def slice_sampler_step_out(log_pivot: float, scale: float,
        sliced_log_density: Callable[[float], float], random_state: np.random.RandomState) -> Tuple[float, float]:

    r = random_state.rand()
    lower_bound = -r * scale
    upper_bound = lower_bound + scale

    def bound_step_out(bound, direction):
        """direction -1 for lower bound, +1 for upper bound"""
        for _ in range(MAX_STEP_OUT):
            if sliced_log_density(bound) <= log_pivot:
                return bound
            else:
                bound += direction * scale
        raise SliceException("Reach maximum iteration ({}) while stepping out for bound ({})".format(MAX_STEP_OUT, direction))

    lower_bound = bound_step_out(lower_bound, -1.)
    upper_bound = bound_step_out(upper_bound, 1.)
    return lower_bound, upper_bound


def slice_sampler_step_in(lower_bound: float, upper_bound: float, log_pivot: float,
        sliced_log_density: Callable[[float], float], random_state: np.random.RandomState) -> float:
    """Find the right amount of movement along with a random_direction"""
    for _ in range(MAX_STEP_LOOP):
        movement = (upper_bound - lower_bound) * random_state.rand() + lower_bound
        if movement == 0.0:
            raise SliceException("The interval for slice sampling has reduced to zero in step in")
        if sliced_log_density(movement) > log_pivot:
            return movement
        else:
            lower_bound = movement if movement < 0.0 else lower_bound
            upper_bound = movement if movement > 0.0 else upper_bound
    raise SliceException("Reach maximum iteration ({}) while stepping in".format(MAX_STEP_LOOP))



import logging

from sklearn.linear_model import LogisticRegression, Ridge, Lasso

from .....ml.constants import BINARY
from .....ml.constants import REGRESSION

L1 = 'L1'
L2 = 'L2'
IGNORE = 'ignore'
ONLY = 'only'
INCLUDE = 'include'

logger = logging.getLogger(__name__)


def get_param_baseline():
    default_params = {
        'C': 1,
        'vectorizer_dict_size': 75000,  # size of TFIDF vectorizer dictionary; used only in text model
        'proc.ngram_range': (1, 5),  # range of n-grams for TFIDF vectorizer dictionary; used only in text model
        'proc.skew_threshold': 0.99,  # numerical features whose absolute skewness is greater than this receive special power-transform preprocessing. Choose big value to avoid using power-transforms
        'proc.impute_strategy': 'median',  # strategy argument of sklearn.SimpleImputer() used to impute missing numeric values
        'penalty': L2,  # regularization to use with regression models
        'handle_text': IGNORE, # how text should be handled: `ignore` - don't use NLP features; `only` - only use NLP features; `include` - use both regular and NLP features
    }
    return default_params


def get_model_params(problem_type: str, hyperparameters):
    penalty = hyperparameters.get('penalty', L2)
    handle_text = hyperparameters.get('handle_text', IGNORE)
    if problem_type == REGRESSION:
        if penalty == L2:
            model_class = Ridge
        elif penalty == L1:
            model_class = Lasso
        else:
            logger.warning('Unknown value for penalty {} - supported types are [l1, l2] - falling back to l2'.format(penalty))
            penalty = L2
            model_class = Ridge
    else:
        model_class = LogisticRegression

    return model_class, penalty, handle_text


def get_default_params(problem_type: str, penalty: str):
    # TODO: get seed from seeds provider
    if problem_type == REGRESSION:
        default_params = {'C': None, 'random_state': 0, 'fit_intercept': True}
        if penalty == L2:
            default_params['solver'] = 'auto'
    else:
        default_params = {'C': None, 'random_state': 0, 'solver': _get_solver(problem_type), 'n_jobs': -1, 'fit_intercept': True}
    model_params = list(default_params.keys())
    return model_params, default_params


def _get_solver(problem_type):
    if problem_type == BINARY:
        # TODO explore using liblinear for smaller datasets
        solver = 'lbfgs'
    else:
        solver = 'lbfgs'
    return solver



import copy
import json
import logging
import os
import random
import time
import warnings
from collections import OrderedDict

import numpy as np
import pandas as pd
from numpy import corrcoef
from pandas import DataFrame, Series
from sklearn.metrics import accuracy_score, balanced_accuracy_score, matthews_corrcoef, f1_score, classification_report  # , roc_curve, auc
from sklearn.metrics import mean_absolute_error, explained_variance_score, r2_score, mean_squared_error, median_absolute_error  # , max_error

from ..constants import BINARY, MULTICLASS, REGRESSION
from ..trainer.abstract_trainer import AbstractTrainer
from ..tuning.ensemble_selection import EnsembleSelection
from ..utils import get_pred_from_proba, get_leaderboard_pareto_frontier, infer_problem_type, augment_rare_classes
from ...data.label_cleaner import LabelCleaner, LabelCleanerMulticlassToBinary
from ...features.generators import PipelineFeatureGenerator
from ...utils.loaders import load_pkl, load_pd
from ...utils.savers import save_pkl, save_pd, save_json
from ...metrics.classification_metrics import confusion_matrix


logger = logging.getLogger(__name__)


# TODO: - Semi-supervised learning
# TODO: - Minimize memory usage of DataFrames (convert int64 -> uint8 when possible etc.)
# Learner encompasses full problem, loading initial data, feature generation, model training, model prediction
# TODO: Loading learner from S3 on Windows may cause issues due to os.path.sep
class AbstractLearner:
    learner_file_name = 'learner.pkl'
    learner_info_name = 'info.pkl'
    learner_info_json_name = 'info.json'

    def __init__(self, path_context: str, label: str, id_columns: list, feature_generator: PipelineFeatureGenerator, label_count_threshold=10,
                 problem_type=None, eval_metric=None, stopping_metric=None, is_trainer_present=False, random_seed=0):
        self.path, self.model_context, self.save_path = self.create_contexts(path_context)
        self.label = label
        self.id_columns = id_columns
        self.threshold = label_count_threshold
        self.problem_type = problem_type
        self.eval_metric = eval_metric
        self.stopping_metric = stopping_metric
        self.is_trainer_present = is_trainer_present
        if random_seed is None:
            random_seed = random.randint(0, 1000000)
        self.random_seed = random_seed
        self.cleaner = None
        self.label_cleaner: LabelCleaner = None
        self.feature_generator: PipelineFeatureGenerator = feature_generator
        self.feature_generators = [self.feature_generator]

        self.trainer: AbstractTrainer = None
        self.trainer_type = None
        self.trainer_path = None
        self.reset_paths = False

        try:
            from .....version import __version__
            self.version = __version__
        except:
            self.version = None

    @property
    def class_labels(self):
        return self.label_cleaner.ordered_class_labels

    def set_contexts(self, path_context):
        self.path, self.model_context, self.save_path = self.create_contexts(path_context)

    def create_contexts(self, path_context):
        model_context = path_context + 'models' + os.path.sep
        save_path = path_context + self.learner_file_name
        return path_context, model_context, save_path

    def fit(self, X: DataFrame, X_val: DataFrame = None, **kwargs):
        self._validate_fit_input(X=X, X_val=X_val, **kwargs)
        return self._fit(X=X, X_val=X_val, **kwargs)

    def _fit(self, X: DataFrame, X_val: DataFrame = None, scheduler_options=None, hyperparameter_tune=False,
            feature_prune=False, holdout_frac=0.1, hyperparameters=None, verbosity=2):
        raise NotImplementedError

    def predict_proba(self, X: DataFrame, model=None, as_pandas=False, as_multiclass=False, inverse_transform=True):
        if as_pandas:
            X_index = copy.deepcopy(X.index)
        else:
            X_index = None
        y_pred_proba = self.load_trainer().predict_proba(self.transform_features(X), model=model)
        if inverse_transform:
            y_pred_proba = self.label_cleaner.inverse_transform_proba(y_pred_proba)
        if as_multiclass and (self.problem_type == BINARY):
            y_pred_proba = LabelCleanerMulticlassToBinary.convert_binary_proba_to_multiclass_proba(y_pred_proba)
        if as_pandas:
            if self.problem_type == MULTICLASS or (as_multiclass and self.problem_type == BINARY):
                y_pred_proba = pd.DataFrame(data=y_pred_proba, columns=self.class_labels, index=X_index)
            else:
                y_pred_proba = pd.Series(data=y_pred_proba, name=self.label, index=X_index)
        return y_pred_proba

    def predict(self, X: DataFrame, model=None, as_pandas=False):
        if as_pandas:
            X_index = copy.deepcopy(X.index)
        else:
            X_index = None
        y_pred_proba = self.predict_proba(X=X, model=model, inverse_transform=False)
        problem_type = self.label_cleaner.problem_type_transform or self.problem_type
        y_pred = get_pred_from_proba(y_pred_proba=y_pred_proba, problem_type=problem_type)
        y_pred = self.label_cleaner.inverse_transform(pd.Series(y_pred))
        if as_pandas:
            y_pred.index = X_index
            y_pred.name = self.label
        else:
            y_pred = y_pred.values
        return y_pred

    def _validate_fit_input(self, X: DataFrame, **kwargs):
        if self.label not in X.columns:
            raise KeyError(f"Label column '{self.label}' is missing from training data. Training data columns: {list(X.columns)}")

    def get_inputs_to_stacker(self, dataset=None, model=None, base_models: list = None, use_orig_features=True):
        if model is not None or base_models is not None:
            if model is not None and base_models is not None:
                raise AssertionError('Only one of `model`, `base_models` is allowed to be set.')

        trainer = self.load_trainer()
        if dataset is None:
            if trainer.bagged_mode:
                dataset_preprocessed = trainer.load_X_train()
                fit = True
            else:
                dataset_preprocessed = trainer.load_X_val()
                fit = False
        else:
            dataset_preprocessed = self.transform_features(dataset)
            fit = False
        if base_models is not None:
            dataset_preprocessed = trainer.get_inputs_to_stacker(X=dataset_preprocessed, base_models=base_models, fit=fit, use_orig_features=use_orig_features)
        elif model is not None:
            base_models = list(trainer.model_graph.predecessors(model))
            dataset_preprocessed = trainer.get_inputs_to_stacker(X=dataset_preprocessed, base_models=base_models, fit=fit, use_orig_features=use_orig_features)
            # Note: Below doesn't quite work here because weighted_ensemble has unique input format returned that isn't a DataFrame.
            # dataset_preprocessed = trainer.get_inputs_to_model(model=model_to_get_inputs_for, X=dataset_preprocessed, fit=fit)

        return dataset_preprocessed

    # Fits _FULL models and links them in the stack so _FULL models only use other _FULL models as input during stacking
    # If model is specified, will fit all _FULL models that are ancestors of the provided model, automatically linking them.
    # If no model is specified, all models are refit and linked appropriately.
    def refit_ensemble_full(self, model='all'):
        return self.load_trainer().refit_ensemble_full(model=model)

    def fit_transform_features(self, X, y=None):
        for feature_generator in self.feature_generators:
            X = feature_generator.fit_transform(X, y)
        return X

    def transform_features(self, X):
        for feature_generator in self.feature_generators:
            X = feature_generator.transform(X)
        return X

    def score(self, X: DataFrame, y=None, model=None):
        if y is None:
            X, y = self.extract_label(X)
        X = self.transform_features(X)
        y = self.label_cleaner.transform(y)
        trainer = self.load_trainer()
        if self.problem_type == MULTICLASS:
            y = y.fillna(-1)
            if (not trainer.eval_metric_expects_y_pred) and (-1 in y.unique()):
                # log_loss / pac_score
                raise ValueError(f'Multiclass scoring with eval_metric=\'{self.eval_metric.name}\' does not support unknown classes.')
        return trainer.score(X=X, y=y, model=model)

    # Scores both learner and all individual models, along with computing the optimal ensemble score + weights (oracle)
    def score_debug(self, X: DataFrame, y=None, extra_info=False, compute_oracle=False, silent=False):
        leaderboard_df = self.leaderboard(extra_info=extra_info, silent=silent)
        if y is None:
            X, y = self.extract_label(X)
        X = self.transform_features(X)
        y = self.label_cleaner.transform(y)
        trainer = self.load_trainer()
        if self.problem_type == MULTICLASS:
            y = y.fillna(-1)
            if (not trainer.eval_metric_expects_y_pred) and (-1 in y.unique()):
                # log_loss / pac_score
                raise ValueError(f'Multiclass scoring with eval_metric=\'{self.eval_metric.name}\' does not support unknown classes.')

        scores = {}
        all_trained_models = trainer.get_model_names_all()
        all_trained_models_can_infer = trainer.get_model_names_all(can_infer=True)
        all_trained_models_original = all_trained_models.copy()
        model_pred_proba_dict, pred_time_test_marginal = trainer.get_model_pred_proba_dict(X=X, models=all_trained_models_can_infer, fit=False, record_pred_time=True)

        if compute_oracle:
            pred_probas = list(model_pred_proba_dict.values())
            ensemble_selection = EnsembleSelection(ensemble_size=100, problem_type=trainer.problem_type, metric=self.eval_metric)
            ensemble_selection.fit(predictions=pred_probas, labels=y, identifiers=None)
            oracle_weights = ensemble_selection.weights_
            oracle_pred_time_start = time.time()
            oracle_pred_proba_norm = [pred * weight for pred, weight in zip(pred_probas, oracle_weights)]
            oracle_pred_proba_ensemble = np.sum(oracle_pred_proba_norm, axis=0)
            oracle_pred_time = time.time() - oracle_pred_time_start
            model_pred_proba_dict['oracle_ensemble'] = oracle_pred_proba_ensemble
            pred_time_test_marginal['oracle_ensemble'] = oracle_pred_time
            all_trained_models.append('oracle_ensemble')

        for model_name, pred_proba in model_pred_proba_dict.items():
            if (trainer.problem_type == BINARY) and (self.problem_type == MULTICLASS):
                pred_proba = self.label_cleaner.inverse_transform_proba(pred_proba)  # FIXME: I think this doesn't work correctly, must use original y as well!
            if trainer.eval_metric_expects_y_pred:
                pred = get_pred_from_proba(y_pred_proba=pred_proba, problem_type=self.problem_type)
                scores[model_name] = self.eval_metric(y, pred)
            else:
                scores[model_name] = self.eval_metric(y, pred_proba)

        pred_time_test = {}
        # TODO: Add support for calculating pred_time_test_full for oracle_ensemble, need to copy graph from trainer and add oracle_ensemble to it with proper edges.
        for model in model_pred_proba_dict.keys():
            if model in all_trained_models_original:
                base_model_set = trainer.get_minimum_model_set(model)
                if len(base_model_set) == 1:
                    pred_time_test[model] = pred_time_test_marginal[base_model_set[0]]
                else:
                    pred_time_test_full_num = 0
                    for base_model in base_model_set:
                        pred_time_test_full_num += pred_time_test_marginal[base_model]
                    pred_time_test[model] = pred_time_test_full_num
            else:
                pred_time_test[model] = None

        scored_models = list(scores.keys())
        for model in all_trained_models:
            if model not in scored_models:
                scores[model] = None
                pred_time_test[model] = None
                pred_time_test_marginal[model] = None

        logger.debug('Model scores:')
        logger.debug(str(scores))
        model_names_final = list(scores.keys())
        df = pd.DataFrame(
            data={
                'model': model_names_final,
                'score_test': list(scores.values()),
                'pred_time_test': [pred_time_test[model] for model in model_names_final],
                'pred_time_test_marginal': [pred_time_test_marginal[model] for model in model_names_final],
            }
        )

        df_merged = pd.merge(df, leaderboard_df, on='model', how='left')
        df_merged = df_merged.sort_values(by=['score_test', 'pred_time_test', 'score_val', 'pred_time_val', 'model'], ascending=[False, True, False, True, False]).reset_index(drop=True)
        df_columns_lst = df_merged.columns.tolist()
        explicit_order = [
            'model',
            'score_test',
            'score_val',
            'pred_time_test',
            'pred_time_val',
            'fit_time',
            'pred_time_test_marginal',
            'pred_time_val_marginal',
            'fit_time_marginal',
            'stack_level',
            'can_infer',
        ]
        df_columns_other = [column for column in df_columns_lst if column not in explicit_order]
        df_columns_new = explicit_order + df_columns_other
        df_merged = df_merged[df_columns_new]

        return df_merged

    def _remove_missing_labels(self, y_true, y_pred):
        """Removes missing labels and produces warning if any are found."""
        if self.problem_type == REGRESSION:
            non_missing_boolean_mask = [y is not None and not pd.isnull(y) for y in y_true]
        else:
            non_missing_boolean_mask = [y is not None and not pd.isnull(y) and y != '' for y in y_true]

        n_missing = len(non_missing_boolean_mask) - sum(non_missing_boolean_mask)
        if n_missing > 0:
            y_true = y_true[non_missing_boolean_mask]
            y_pred = y_pred[non_missing_boolean_mask]
            warnings.warn(f"There are {n_missing} (out of {len(y_true)}) evaluation datapoints for which the label is missing. "
                          f"AutoGluon removed these points from the evaluation, which thus may not be entirely representative. "
                          f"You should carefully study why there are missing labels in your evaluation data.")
        return y_true, y_pred

    # TODO: Refactor to be less brittle.
    # TODO: Instead take y_pred_proba as input, convert to y_pred when necessary. Treat y_pred_proba as y_pred for regression.
    #  This makes this function behave much nicer, currently confusion matrix is only able to be constructed when y_pred is supplied, and crashes when y_pred_proba is supplied.
    #  Therefore, it is impossible to get confusion matrix when eval_metric is log_loss. This shouldn't be the case.
    def evaluate(self, y_true, y_pred, silent=False, auxiliary_metrics=False, detailed_report=True, high_always_good=False):
        """ Evaluate predictions.
            Args:
                silent (bool): Should we print which metric is being used as well as performance.
                auxiliary_metrics (bool): Should we compute other (problem_type specific) metrics in addition to the default metric?
                detailed_report (bool): Should we computed more-detailed versions of the auxiliary_metrics? (requires auxiliary_metrics=True).
                high_always_good (bool): If True, this means higher values of returned metric are ALWAYS superior (so metrics like MSE should be returned negated)

            Returns single performance-value if auxiliary_metrics=False.
            Otherwise returns dict where keys = metrics, values = performance along each metric.
        """
        assert isinstance(y_true, (np.ndarray, pd.Series))
        assert isinstance(y_pred, (np.ndarray, pd.Series))  # TODO: Enable DataFrame for y_pred_proba

        # TODO: Consider removing _remove_missing_labels, this creates an inconsistency between how .score, .score_debug, and .evaluate compute scores.
        y_true, y_pred = self._remove_missing_labels(y_true, y_pred)
        performance = self.eval_metric(y_true, y_pred)

        metric = self.eval_metric.name

        if not high_always_good:
            performance = performance * self.eval_metric._sign  # flip negative once again back to positive (so higher is no longer necessarily better)

        if not silent:
            logger.log(20, f"Evaluation: {metric} on test data: {performance}")

        if not auxiliary_metrics:
            return performance

        # Otherwise compute auxiliary metrics:
        auxiliary_metrics = []
        if self.problem_type == REGRESSION:  # Adding regression metrics
            pearson_corr = lambda x, y: corrcoef(x, y)[0][1]
            pearson_corr.__name__ = 'pearson_correlation'
            auxiliary_metrics += [
                mean_absolute_error, explained_variance_score, r2_score, pearson_corr, mean_squared_error, median_absolute_error,
                # max_error
            ]
        else:  # Adding classification metrics
            auxiliary_metrics += [accuracy_score, balanced_accuracy_score, matthews_corrcoef]
            if self.problem_type == BINARY:  # binary-specific metrics
                # def auc_score(y_true, y_pred): # TODO: this requires y_pred to be probability-scores
                #     fpr, tpr, _ = roc_curve(y_true, y_pred, pos_label)
                #   return auc(fpr, tpr)
                f1micro_score = lambda y_true, y_pred: f1_score(y_true, y_pred, average='micro')
                f1micro_score.__name__ = f1_score.__name__
                auxiliary_metrics += [f1micro_score]  # TODO: add auc?
            # elif self.problem_type == MULTICLASS:  # multiclass metrics
            #     auxiliary_metrics += []  # TODO: No multi-class specific metrics for now. Include top-5, top-10 accuracy here.

        performance_dict = OrderedDict({metric: performance})
        for metric_function in auxiliary_metrics:
            if isinstance(metric_function, tuple):
                metric_function, metric_kwargs = metric_function
            else:
                metric_kwargs = None
            metric_name = metric_function.__name__
            if metric_name not in performance_dict:
                try:  # only compute auxiliary metrics which do not error (y_pred = class-probabilities may cause some metrics to error)
                    if metric_kwargs:
                        performance_dict[metric_name] = metric_function(y_true, y_pred, **metric_kwargs)
                    else:
                        performance_dict[metric_name] = metric_function(y_true, y_pred)
                except ValueError:
                    pass

        if not silent:
            logger.log(20, "Evaluations on test data:")
            logger.log(20, json.dumps(performance_dict, indent=4))

        if detailed_report and (self.problem_type != REGRESSION):
            # Construct confusion matrix
            try:
                performance_dict['confusion_matrix'] = confusion_matrix(y_true, y_pred, labels=self.label_cleaner.ordered_class_labels, output_format='pandas_dataframe')
            except ValueError:
                pass
            # One final set of metrics to report
            cl_metric = lambda y_true, y_pred: classification_report(y_true, y_pred, output_dict=True)
            metric_name = 'classification_report'
            if metric_name not in performance_dict:
                try:  # only compute auxiliary metrics which do not error (y_pred = class-probabilities may cause some metrics to error)
                    performance_dict[metric_name] = cl_metric(y_true, y_pred)
                except ValueError:
                    pass
                if not silent and metric_name in performance_dict:
                    logger.log(20, "Detailed (per-class) classification report:")
                    logger.log(20, json.dumps(performance_dict[metric_name], indent=4))
        return performance_dict

    def extract_label(self, X):
        if self.label not in list(X.columns):
            raise ValueError(f"Provided DataFrame does not contain label column: {self.label}")
        y = X[self.label].copy()
        X = X.drop(self.label, axis=1)
        return X, y

    def leaderboard(self, X=None, y=None, extra_info=False, only_pareto_frontier=False, silent=False):
        if X is not None:
            leaderboard = self.score_debug(X=X, y=y, extra_info=extra_info, silent=True)
        else:
            trainer = self.load_trainer()
            leaderboard = trainer.leaderboard(extra_info=extra_info)
        if only_pareto_frontier:
            if 'score_test' in leaderboard.columns and 'pred_time_test' in leaderboard.columns:
                score_col = 'score_test'
                inference_time_col = 'pred_time_test'
            else:
                score_col = 'score_val'
                inference_time_col = 'pred_time_val'
            leaderboard = get_leaderboard_pareto_frontier(leaderboard=leaderboard, score_col=score_col, inference_time_col=inference_time_col)
        if not silent:
            with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.width', 1000):
                print(leaderboard)
        return leaderboard

    # TODO: cache_data must be set to True to be able to pass X and y as None in this function, otherwise it will error.
    # Warning: This can take a very, very long time to compute if the data is large and the model is complex.
    # A value of 0.01 means that the objective metric error would be expected to increase by 0.01 if the feature were removed.
    # Negative values mean the feature is likely harmful.
    # model: model (str) to get feature importances for, if None will choose best model.
    # features: list of feature names that feature importances are calculated for and returned, specify None to get all feature importances.
    # feature_stage: Whether to compute feature importance on raw original features ('original'), transformed features ('transformed') or on the features used by the particular model ('transformed_model').
    def get_feature_importance(self, model=None, X=None, y=None, features: list = None, feature_stage='original', subsample_size=1000, silent=False) -> Series:
        valid_feature_stages = ['original', 'transformed', 'transformed_model']
        if feature_stage not in valid_feature_stages:
            raise ValueError(f'feature_stage must be one of: {valid_feature_stages}, but was {feature_stage}.')
        trainer = self.load_trainer()
        if X is not None:
            if y is None:
                X, y = self.extract_label(X)
            y = self.label_cleaner.transform(y)
            X, y = self._remove_nan_label_rows(X, y)

            if feature_stage == 'original':
                return trainer._get_feature_importance_raw(model=model, X=X, y=y, features_to_use=features, subsample_size=subsample_size, transform_func=self.transform_features, silent=silent)
            X = self.transform_features(X)
        else:
            if feature_stage == 'original':
                raise AssertionError('Feature importance `dataset` cannot be None if `feature_stage==\'original\'`. A test dataset must be specified.')
            y = None
        raw = feature_stage == 'transformed'
        return trainer.get_feature_importance(X=X, y=y, model=model, features=features, raw=raw, subsample_size=subsample_size, silent=silent)

    @staticmethod
    def _remove_nan_label_rows(X, y):
        if y.isnull().any():
            y = y.dropna()
            X = X.loc[y.index]
        return X, y

    @staticmethod
    def infer_problem_type(y: Series):
        return infer_problem_type(y=y)

    def save(self):
        trainer = None
        if self.trainer is not None:
            if not self.is_trainer_present:
                self.trainer.save()
                trainer = self.trainer
                self.trainer = None
        save_pkl.save(path=self.save_path, object=self)
        self.trainer = trainer

    # reset_paths=True if the learner files have changed location since fitting.
    # TODO: Potentially set reset_paths=False inside load function if it is the same path to avoid re-computing paths on all models
    # TODO: path_context -> path for v0.1
    @classmethod
    def load(cls, path_context, reset_paths=True):
        load_path = path_context + cls.learner_file_name
        obj = load_pkl.load(path=load_path)
        if reset_paths:
            obj.set_contexts(path_context)
            obj.trainer_path = obj.model_context
            obj.reset_paths = reset_paths
            # TODO: Still have to change paths of models in trainer + trainer object path variables
            return obj
        else:
            obj.set_contexts(obj.path_context)
            return obj

    def save_trainer(self, trainer):
        if self.is_trainer_present:
            self.trainer = trainer
            self.save()
        else:
            self.trainer_path = trainer.path
            trainer.save()

    def load_trainer(self) -> AbstractTrainer:
        if self.trainer is not None:
            return self.trainer
        else:
            return self.trainer_type.load(path=self.trainer_path, reset_paths=self.reset_paths)

    # Loads models in memory so that they don't have to be loaded during predictions
    def persist_trainer(self, low_memory=False, models='all', with_ancestors=False, max_memory=None) -> list:
        self.trainer = self.load_trainer()
        if not low_memory:
            return self.trainer.persist_models(models, with_ancestors=with_ancestors, max_memory=max_memory)
            # Warning: After calling this, it is not necessarily safe to save learner or trainer anymore
            #  If neural network is persisted and then trainer or learner is saved, there will be an exception thrown
        else:
            return []

    def distill(self, X=None, y=None, X_val=None, y_val=None, time_limits=None, hyperparameters=None, holdout_frac=None,
                verbosity=None, models_name_suffix=None, teacher_preds='soft',
                augmentation_data=None, augment_method='spunge', augment_args={'size_factor':5,'max_size':int(1e5)}):
        """ See abstract_trainer.distill() for details. """
        if X is not None:
            if (self.eval_metric is not None) and (self.eval_metric.name == 'log_loss') and (self.problem_type == MULTICLASS):
                X = augment_rare_classes(X, self.label, self.threshold)
            if y is None:
                X, y = self.extract_label(X)
            X = self.transform_features(X)
            y = self.label_cleaner.transform(y)
            if self.problem_type == MULTICLASS:
                y = y.fillna(-1)
        else:
            y = None

        if X_val is not None:
            if X is None:
                raise ValueError("Cannot specify X_val without specifying X")
            if y_val is None:
                X_val, y_val = self.extract_label(X_val)
            X_val = self.transform_features(X_val)
            y_val = self.label_cleaner.transform(y_val)

        if augmentation_data is not None:
            augmentation_data = self.transform_features(augmentation_data)

        trainer = self.load_trainer()
        distilled_model_names = trainer.distill(X_train=X, y_train=y, X_val=X_val, y_val=y_val, time_limits=time_limits, hyperparameters=hyperparameters,
                                                holdout_frac=holdout_frac, verbosity=verbosity, teacher_preds=teacher_preds, models_name_suffix=models_name_suffix,
                                                augmentation_data=augmentation_data, augment_method=augment_method, augment_args=augment_args)
        self.save_trainer(trainer=trainer)
        return distilled_model_names

    @classmethod
    def load_info(cls, path, reset_paths=True, load_model_if_required=True):
        load_path = path + cls.learner_info_name
        try:
            return load_pkl.load(path=load_path)
        except Exception as e:
            if load_model_if_required:
                learner = cls.load(path_context=path, reset_paths=reset_paths)
                return learner.get_info()
            else:
                raise e

    def save_info(self, include_model_info=False):
        info = self.get_info(include_model_info=include_model_info)

        save_pkl.save(path=self.path + self.learner_info_name, object=info)
        save_json.save(path=self.path + self.learner_info_json_name, obj=info)
        return info

    # TODO: Add data info gathering at beginning of .fit() that is used by all learners to add to get_info output
    # TODO: Add feature inference / feature engineering info to get_info output
    def get_info(self, **kwargs):
        learner_info = {
            'path': self.path,
            'label': self.label,
            'random_seed': self.random_seed,
            'version': self.version,
        }

        return learner_info
